#version 330
uniform sampler2D videoTexture;
uniform sampler2D inputFrameTexture;
out vec4 FragColour;
in vec2 out_uvs;

uniform int mode;
uniform int preserveAlphaL;
uniform int unmult;
uniform float offsetHue;
uniform float strengthF;

uniform int iterationsL;
uniform float aspectF;
uniform vec2 resV;
uniform float offsetDistort;

const float eps = 0.0000001;

vec3 hsv2rgb( in vec3 c )
{
    vec3 rgb = clamp( abs(mod(c.x*6.0+vec3(0.0,4.0,2.0),6.0)-3.0)-1.0, 0.0, 1.0 );
    return c.z * mix( vec3(1.0), rgb, c.y);
}

vec3 rgb2hsv( in vec3 c)
{
    vec4 k = vec4(0.0, -1.0/3.0, 2.0/3.0, -1.0);
    vec4 p = mix(vec4(c.zy, k.wz), vec4(c.yz, k.xy), (c.z<c.y) ? 1.0 : 0.0);
    vec4 q = mix(vec4(p.xyw, c.x), vec4(c.x, p.yzx), (p.x<c.x) ? 1.0 : 0.0);
    float d = q.x - min(q.w, q.y);
    return vec3(abs(q.z + (q.w - q.y) / (6.0*d+eps)), d / (q.x+eps), q.x);
}

vec3 rgb2hsl(vec3 col)
{
    float minc = min( col.r, min(col.g, col.b) );
    float maxc = max( col.r, max(col.g, col.b) );
    vec3  mask = step(col.grr,col.rgb) * step(col.bbg,col.rgb);
    vec3 h = mask * (vec3(0.0,2.0,4.0) + (col.gbr-col.brg)/(maxc-minc + eps)) / 6.0;
    return vec3( fract( 1.0 + h.x + h.y + h.z ),              // H
                 (maxc-minc)/(1.0-abs(minc+maxc-1.0) + eps),  // S
                 (minc+maxc)*0.5 );                           // L
}

vec3 hsl2rgb(in vec3 c)
{
    vec3 rgb = clamp( abs(mod(c.x*6.0+vec3(0.0,4.0,2.0),6.0)-3.0)-1.0, 0.0, 1.0 );
    return c.z + c.y * (rgb-0.5)*(1.0-abs(2.0*c.z-1.0));
}

vec2 distortUV(vec2 uv, float k, float kcube)
{
    k /= 1000.0;
    k *= min(resV.x, resV.y); // normalise based on res

    vec2 t = uv - .50;
    float r2 = (t.x * t.x / aspectF) + (t.y * t.y * aspectF);
    float f = 0.;

    f = 1. + r2 * k;
    
    vec2 nUv = f * t;

    nUv += .50;
    
    return nUv;
}
                             
// postP
void main(void)
{
    // splice in original
    vec2 input_coords = out_uvs;
    vec4 aberratedFrag = texture(videoTexture, out_uvs); // output from aberrate.frag,
    // this has the first aberrated channel in the red and it's alpha in green
    // the second aberrated channel is in the blue and alpha
    
    vec2 uvmod = out_uvs;
    vec4 inputRGBA = vec4(0.0);
    
    // distort
    float iterationsL = float(iterationsL);
    float iterationMultF = 1.0 / iterationsL;
    
    float uniformDistortStart = offsetDistort * 0.9;
    float offsetDistortUniformIncrF = iterationMultF * (offsetDistort - uniformDistortStart);

    // apply uniform distortion to original channel if applicable
    if(offsetDistort != 0.0)
    {
        for(float i = 1; i <= iterationsL; i++)
        {
            uvmod = distortUV(out_uvs, (uniformDistortStart + (offsetDistortUniformIncrF * i)), 0.0);
            inputRGBA += texture(inputFrameTexture, uvmod);
        }
        
        inputRGBA *= vec4(iterationMultF);
    }
    else
    {
        inputRGBA = texture(inputFrameTexture, uvmod);
    }
    // end distort

    FragColour = vec4(0.0);
    
    float red, redAlpha, blue, blueAlpha, green, greenAlpha, screenAlpha;
    
    // alpha mix with just the unchanged alpha
    if(mode == 1) // red & blue
    {
        red = aberratedFrag.r;
        redAlpha = aberratedFrag.g;
        
        blue = aberratedFrag.b;
        blueAlpha = aberratedFrag.a;

        // if preserve alpha is enabled, multiply colour channels by their own alpha channels. otherwise the colour channels are already premult
        if(preserveAlphaL == 0)
        {
            // create our own alpha by screening all other alpha's
            screenAlpha = inputRGBA.a + ((1.0 - inputRGBA.a) * blueAlpha);
            screenAlpha += (1.0 - screenAlpha) * redAlpha;
            // now that we have the result of the screen alpha, they aren't needed
            redAlpha = 1.0; // don't mult colour channel by it's alpha because it's already premult
            blueAlpha = 1.0;
        }
        
        FragColour.rgb = vec3(red * redAlpha, inputRGBA.g, blue * blueAlpha);
    }
    else if(mode == 2) // red & green
    {
        red = aberratedFrag.r;
        redAlpha = aberratedFrag.g;
        
        green = aberratedFrag.b;
        greenAlpha = aberratedFrag.a;
        
        // if preserve alpha is enabled, multiply colour channels by their own alpha channels. otherwise the colour channels are already premult
        if(preserveAlphaL == 0)
        {
            // create our own alpha by screening all other alpha's
            screenAlpha = inputRGBA.a + ((1.0 - inputRGBA.a) * redAlpha);
            screenAlpha += (1.0 - screenAlpha) * greenAlpha;
            // now that we have the result of the screen alpha, they aren't needed
            redAlpha = 1.0; // don't mult colour channel by it's alpha because it's already premult
            greenAlpha = 1.0;
        }
        
        FragColour.rgb = vec3(red * redAlpha, green * greenAlpha, inputRGBA.b);
    }
    else // green & blue
    {
        green = aberratedFrag.r;
        greenAlpha = aberratedFrag.g;
        
        blue = aberratedFrag.b;
        blueAlpha = aberratedFrag.a;
        
        // if preserve alpha is enabled, multiply colour channels by their own alpha channels. otherwise the colour channels are already premult
        if(preserveAlphaL == 0)
        {
            // create our own alpha by screening all other alpha's
            screenAlpha = inputRGBA.a + ((1.0 - inputRGBA.a) * greenAlpha);
            screenAlpha += (1.0 - screenAlpha) * blueAlpha;
            // now that we have the result of the screen alpha, they aren't needed
            blueAlpha = 1.0; // don't mult colour channel by it's alpha because it's already premult
            greenAlpha = 1.0;
        }
        
        FragColour.rgb = vec3(inputRGBA.r, green * greenAlpha, blue * blueAlpha);
    }

    if(preserveAlphaL == 1)
    {
        FragColour.a = texture(inputFrameTexture, out_uvs).a; // original alpha
    }
    else
    {
        FragColour.a = mix(screenAlpha, inputRGBA.a, strengthF);
    }

    // unmult if requested
    float alpha = 0.0;
    
    if(unmult == 1)
    {
        alpha = max(FragColour.r, FragColour.g);
        alpha = max(alpha, FragColour.b);
        
        if(preserveAlphaL == 1)
        {
            alpha *= FragColour.a;
        }
        
        FragColour.a = min(alpha, 1.0); // assign and clamp between 0 to 1
    }
    
    // mix with original
    if(strengthF > 0.0)
    {
        if(mode == 1 || mode == 2)
        {
            FragColour.r = mix(FragColour.r, inputRGBA.r, strengthF);
        }
        
        if(mode == 2 || mode == 3)
        {
            FragColour.g = mix(FragColour.g, inputRGBA.g, strengthF);
        }
        
        if(mode == 1 || mode == 3)
        {
            FragColour.b = mix(FragColour.b, inputRGBA.b, strengthF);
        }
    }
    
    FragColour.a = mix(FragColour.a, inputRGBA.a, strengthF);
    
    if(offsetHue != 0.0)
    {
        FragColour.rgb = rgb2hsv(FragColour.rgb);
        FragColour.r += offsetHue;
        FragColour.rgb = hsv2rgb(FragColour.rgb);
    }
    
    FragColour = FragColour.argb; // swizzle

    //FragColour = texture(videoTexture, out_uvs); // debug, this shader doesn't do anything
}
